import re
from datetime import datetime
from io import StringIO
import os
from pathlib import Path
from typing import Mapping, Union

import pandas as pd
from tqdm import tqdm


# Regex to match the line
_RE_LINE_PATTERN = re.compile(
    r"(?P<permissions>[\w-]{10})\s+"
    r"(?P<hardlinks>\d+)\s+"
    r"(?P<owner>\S+)\s+"
    r"(?P<group>\S+)\s+"
    r"(?P<size>\d+)\s+"
    r"(?P<month>\w{3})\s+"
    r"(?P<day>\d{1,2})\s+"
    r"(?P<time_or_year>\d{2}:\d{2}|\d{4})\s+"
    r"(?P<filename>.+?)"
    r"(?:\s+->\s+(?P<symlink_target>.+))?$"
)


def _parse_line(line: str):
    # pylint: disable=W0718

    match = _RE_LINE_PATTERN.match(line.strip())
    if not match:
        return None
    parts = match.groupdict()

    # Parse datetime
    month = parts["month"]
    day = parts["day"]
    time_or_year = parts["time_or_year"]
    try:
        if ":" in time_or_year:
            # No year in output, use current year
            dt_str = f"{month} {day} {datetime.now().year} {time_or_year}"
            dt = datetime.strptime(dt_str, "%b %d %Y %H:%M")
        else:
            dt_str = f"{month} {day} {time_or_year}"
            dt = datetime.strptime(dt_str, "%b %d %Y")
    except Exception:
        dt = pd.NaT

    return {
        "Permissions": parts["permissions"],
        "Hardlinks": int(parts["hardlinks"]),
        "Owner": parts["owner"],
        "Group": parts["group"],
        "Filesize": int(parts["size"]),
        "Datetime": dt,
        "Filename": parts["filename"].strip(),
        "SymlinkTarget": parts["symlink_target"],
        "IsSymlink": parts["permissions"].startswith("l"),
    }


def _parse_iterable(iterable, keep_dots):
    records = []
    for line in iterable:
        parsed = _parse_line(line)
        if parsed:
            records.append(parsed)
    df = pd.DataFrame(records)
    if not keep_dots:
        index_dots = df[(df["Filename"] == ".") | (df["Filename"] == "..")].index
        df.drop(index_dots, inplace=True)
    return df


def read_ls(file: Union[str, Path, StringIO], keep_dots=False) -> pd.DataFrame:
    """
    Reads the output of an `ls -l`-like command and returns a DataFrame.
    Handles symlinks and combines Month, Day, Time/Year into a single Datetime column.
    """
    if isinstance(file, (str, Path)):
        with open(file, "r", encoding="utf-8") as f:
            next(f, None)  # Skip header
            return _parse_iterable(f, keep_dots).sort_values(
                "Datetime", ignore_index=True
            )

    elif isinstance(file, StringIO):
        return _parse_iterable(file, keep_dots).sort_values(
            "Datetime", ignore_index=True
        )

    else:
        raise TypeError(
            f"'file' must be str, Path, or StringIO, not {type(file).__name__}"
        )


def _join_filetree_dict(ls_dict: Mapping[Path, pd.DataFrame], common: bool):
    dfs = []
    for path, path_df in ls_dict.items():
        path_df["Parent"] = str(path)
        dfs.append(path_df)

    df = pd.concat(dfs)

    if common:
        value = os.path.commonprefix(list(df["Parent"]))
        df["Parent"] = df["Parent"].apply(
            lambda x: x.removeprefix(value).removeprefix("/")
        )

    return df


def _process_file(path: str | Path):
    chunks = {}
    current_io = StringIO()
    current_name = None
    with open(path, "r", encoding="utf-8") as rf:
        for i, line in enumerate(rf):
            if line.strip() == "":
                if current_name is not None:
                    chunks[current_name] = current_io
                    current_name = None
                    current_io = StringIO()
                continue
            if line.startswith("====="):
                current_name = line.strip().strip("=").strip()
                next(rf)
                next(rf)
                continue
            current_io.write(line)
    return chunks


def _process_file_chunks(chunks: Mapping[Path, StringIO], keep_dots):
    # pylint: disable=W0718
    ls_dict = {}
    for fpath, io in tqdm(chunks.items()):
        try:
            ls_dict[fpath] = read_ls(StringIO(io.getvalue()), keep_dots)
        except Exception as exc:
            raise Exception(f"Error reading {fpath}: {str(exc)}") from exc
    return ls_dict


def read_filetree_ls(
    file: str | Path,
    join: bool = False,
    common: bool = True,
    keep_dots: bool = False,
) -> Mapping[Path, pd.DataFrame] | pd.DataFrame:

    chunks = _process_file(file)
    ls_dict = _process_file_chunks(chunks, keep_dots)
    # return chunks
    if join:
        return _join_filetree_dict(ls_dict, common)

    return ls_dict
